from scipy.interpolate import interpn
import numpy as np
from .HeST_Core import CPD_Signal, Random_QPmomentum, QP_dispersion, QP_velocity 
from .HeST_Core import Singlet_PhotonEnergy

class VCPD:
    def __init__(self, surface_condition, baselineNoise, phononConversion=0.25):
        '''
        Create a Virtual CPD class to keep track of individual CPDs
        and add them to the VDetector object.

        Args:
        surface_condition: a function that tracks where photon/QP paths are obstructed by
                          the CPD surface. Returns (True, "CPD") if the photon is unobstructed;
                          Returns (False, "CPD") if the photon path hits the CPD surface.
        baselineNoise: [mean, width] of the CPDs baseline noise
        phononConversion=0.25: fraction of photon energy that creates a phonon signal in the CPD

        '''

        self.surface_condition = surface_condition
        self.baselineNoise     = np.array( baselineNoise )
        self.phononConversion  = phononConversion


    def set_surface_condition(self, f1):
        self.surface_condition = f1
    def set_baselineNoise(self, p1):
        self.baselineNoise = np.array( p1 )
    def set_phononConversion(self, p1):
        self.phononConversion = p1

    def get_surface_condition(self):
        return self.surface_condition
    def get_baselineNoise(self):
        return self.baselineNoise
    def get_phononConversion(self):
        return self.phononConversion

    def check_surface(self, x, y, z):
        return self.surface_condition(x, y, z)



class VDetector:
    def __init__(self, surface_conditions, liquid_surface, liquid_conditions,
                 adsorption_gain, evaporation_eff, CPDs=[], LCEmap=0, LCEmap_positions=0., QPEmap=0, QPEmap_positions=0.,
                 photon_reflection_prob=0., QP_reflection_prob=0.):
        
        '''
        Create a Virtual Detector class.
        
        Args:
        surface_conditions: a list of functions that tracks where photon/QP paths are obstructed by 
                            the Detector surfaces. Each function returns (True, *SurfaceType) if the photon is unobstructed; 
                            Returns (False, *surfaceType) if the photon path hits the detector surface.
                            *SurfaceType is a way of tracking what type of boundary this is: e.g. "X", "Y", "Z", or "XY", making it 
                            easy to track how the particle may reflect off of that boundary
                            
        liquid_surface:     A function that tracks where the liquid surface is, similar to the surface conditions above. This is used in QP propagation/evaporation calculations.
                            The surfaceType here must be "Liquid"
        liquid_conditions:  A function returning true if the X,Y,Z position is inside the LHe volume, and False if outside the volume

        CPDs:               A list of VCPD objects to keep track of
        
        adsorption_gain:    Added energy from adsorption of evaporated QPs, per QP, in eV
        
        self.evaporation_eff:  Flat efficiency factor on QP evaporation, between 0 and 1
        
        LCEmap:             A 3D array tracking positions with mean photon collection probability, dimensionality (M x N x L)
        
        LCEmap_positions:   A set of three 1-D arrays (x, y, z) to associate the LCE map entries in (M x N x L) discrete bins to individual x,y,z coordinates.
                            If loading a premade map, these must match what was used for that map's generator. 
        
        QPEmap:             A 3D array tracking position with mean QP evaporation probability. Different from LCEmap due to liquid surface physics
        
        QPEmap_positions:   A set of three 1-D arrays (x, y, z) to associate the QPE map entries in (M x N x L) discrete bins to individual x,y,z coordinates. 
                            If loading a premade map, these must match what was used for that map's generator. The dimensionality of the QP map doesn't need to
                            match the dimensionality of the LCE map.
                            
        photon_reflection_prob: probability that a photon reflects off of detector surfaces, between 0 and 1
        QP_reflection_prob: probability that a QP reflects off of detector surfaces, between 0 and 1
        '''
        
        self.surface_conditions = surface_conditions
        self.liquid_surface     = liquid_surface
        self.liquid_conditions  = liquid_conditions
        self.CPDs               = CPDs
        self.adsorption_gain    = adsorption_gain #meV
        self.evaporation_eff    = evaporation_eff
        self.LCEmap             = LCEmap
        self.LCEmap_positions   = LCEmap_positions
        self.QPEmap             = QPEmap
        self.QPEmap_positions   = QPEmap_positions
        self.photon_reflection_prob = photon_reflection_prob
        self.QP_reflection_prob = QP_reflection_prob
        
        
    
        
    def set_surface_conditions(self, f1):
        self.surface_conditions = list(f1)
    def add_surface_condition(self, f1):
        self.surface_conditions.append( f1 )
        
    def set_CPDs(self, p1):
        self.CPDs = list(p1)
    def add_CPD(self, p1):
        self.CPDs.append( p1 )
        
    def set_liquid_surface(self, f1):
        self.liquid_surface = f1
    def set_liquid_conditions(self, f1):
        self.liquid_conditions = f1
    def set_adsorption_gain( self, p1 ):
        self.adsorption_gain = p1
    def set_evaporation_eff( self, p1 ):
        self.evaporation_eff = p1
        
        
    def load_LCEmap(self, filename):
        self.LCEmap = np.load(filename)
    def set_LCEmap_positions(self, p1):
        if len(p1) == 3:
            self.LCEmap_positions = p1
        else: 
            print("This function requires a list of 3 individual 1-D arrays to associate the LCE map with X,Y,Z coordinates")
    def load_QPEmap(self, filename):
        self.QPEmap = np.load(filename)
    def set_QPEmap_positions(self, p1):
        if len(p1) == 3:
            self.QPEmap_positions = p1
        else: 
            print("This function requires a list of 3 individual 1-D arrays to associate the QP evaporation map with X,Y,Z coordinates")
    
    def set_photon_reflection_prob(self, p1):
        self.photon_reflection_prob = p1
    def set_QP_reflection_prob(self, p1):
        self.QP_reflection_prob = p1
        
    def get_surface_conditions(self):
        return list(self.surface_conditions)
    def get_liquid_surface(self):
        return self.liquid_surface
    def get_liquid_conditions(self):
        return self.liquid_conditions
    
    def get_LCEmap(self):
        return self.LCEmap
    
    def get_LCEmap_positions(self):
        return self.LCEmap_positions
    
    def get_QPEmap(self):
        return self.LCEmap_positions
    
    def get_QPEmap_positions(self):
        return self.QPEmap_positions
        
        
        
    def get_nCPDs(self):
        return len(self.CPDs)
    
    def get_CPD(self, index):
        return self.CPDs[index]
    
    def get_liquid_surface( self ):
        return self.liquid_surface
    def get_adsorption_gain(self):
        return self.adsorption_gain
    def get_evaporation_eff(self):
        return self.evaporation_eff
    
    def get_photon_reflection_prob(self):
        return self.photon_reflection_prob
    def get_QP_reflection_prob(self):
        return self.QP_reflection_prob
    
    
    def create_LCEmap(self, x_array, y_array, z_array, nPhotons=10000, filestring="detector_LCEmap"):
        print("Creating LCE map for this detector geometry...")
        x, y, z = np.array(x_array), np.array(y_array), np.array(z_array)
        self.set_LCEmap_positions( [x, y, z] )
        nCPDs = self.get_nCPDs()
        m = np.zeros((len(x), len(y), len(z), nCPDs), dtype=float)
        conditions = self.get_surface_conditions()
        for i in range(nCPDs):
            conditions.append( (self.get_CPD(i)).get_surface_condition() )
        
        refl_prob = self.get_photon_reflection_prob()
        for xx in range(len(x)):
            print("%i / %i" % (xx, len(x)))
            for yy in range(len(y)):
                for zz in range(len(z)):
                    
                    pos = np.array([x[xx], y[yy], z[zz]])
                    if (self.get_liquid_conditions())(*pos) == False:
                        continue
                    hitProbs = [0.]*nCPDs
                    for n in range(nPhotons):
                    
                        hit, arrival_time, n, xs, ys, zs, cpd_id = photon_propagation(pos, conditions, refl_prob)
                        if hit > 0:
                            hitProbs[cpd_id] += 1.
                        
                        
                                
                    hitProbs = np.array(hitProbs)/nPhotons
                    
                    m[xx][yy][zz] = hitProbs
                    #print(hitProbs)
                    
        self.LCEmap = m
        np.save(filestring+".npy", m)
        print("Saved map to %s.npy" % filestring)
        
    
        
        
    def get_photon_hits(self, X, Y, Z):
        if type(self.LCEmap) == int:
            print("No LCE Map Loaded!")
            return -999
        
        positions = self.get_LCEmap_positions()
        x = positions[0]
        y = positions[1]
        z = positions[2]
        return interpn((x, y, z), self.LCEmap, [X, Y, Z])[0]
    
    def create_QPEmap(self, x_array, y_array, z_array, nQPs=10000, filestring="detector_QPEmap"):
        print("Creating QPE map for this detector geometry...")
        x, y, z = np.array(x_array), np.array(y_array), np.array(z_array)
        self.set_QPEmap_positions( [x, y, z] )
        nCPDs = self.get_nCPDs()
        m = np.zeros((len(x), len(y), len(z), nCPDs), dtype=float)
        conditions = self.get_surface_conditions()
        conditions.append( self.liquid_surface )
        for i in range(nCPDs):
            conditions.append( (self.get_CPD(i)).get_surface_condition() )
        
        refl_prob = self.get_QP_reflection_prob()
        for xx in range(len(x)):
            print("%i / %i" % (xx, len(x)))
            for yy in range(len(y)):
                for zz in range(len(z)):
                    
                    pos = np.array([x[xx], y[yy], z[zz]])
                    if (self.get_liquid_conditions())(*pos) == False:
                        continue
                    hitProbs = [0.]*nCPDs
                    for n in range(nQPs):

                        hit, time, n, xs, ys, zs, p, surf, cpd_id = QP_propagation(pos, conditions, refl_prob, evap_eff=self.evaporation_eff)
                        if hit > 0:
                            hitProbs[cpd_id] += 1.
                        
                                
                    hitProbs = np.array(hitProbs)/nQPs
                    
                    m[xx][yy][zz] = hitProbs
                    #print(hitProbs)
                    
        self.QPEmap = m
        np.save(filestring+".npy", m)
        print("Saved map to %s.npy" % filestring)
    
    def get_QP_hits(self, X, Y, Z):
        if type(self.QPEmap) == int:
            print("No QPE Map Loaded!")
            return -999
        
        positions = self.get_QPEmap_positions()
        x = positions[0]
        y = positions[1]
        z = positions[2]
        return interpn((x, y, z), self.QPEmap, [X, Y, Z])[0]




''' #############################################################################

    Define various functions to calculate particle propagation through the 
    Detector geometry

    ############################################################################# '''

def find_surface_intersection(start, direction, conditions):
    t = np.linspace(0, 8, 500)  # Parameter range for the line
    # Calculate the line coordinates
    x_line = start[0] + t * direction[0]
    y_line = start[1] + t * direction[1]
    z_line = start[2] + t * direction[2]
    intersection_points = []
    dist = 999.
    coords = [None, None, None]
    surface_type = None
    for cond in conditions:
        cut, surface = cond(x_line, y_line, z_line)
        first_int = np.argmax(~cut)
        if first_int == 0:
            continue
        d = t[first_int]
        if d < dist:
            dist = d
            #get the coords of the first point *before* the interaction
            coords = [x_line[first_int-1], y_line[first_int-1], z_line[first_int-1]]
            surface_type = surface
    return coords[0], coords[1], coords[2], surface_type

def evaporation(momentum, energy, velocity, direction):
    #adapted from Pratyush's code
    mass = 3.725472e9 #He mass in eV/c^2
    c = 2.998e8 #m/s
    Eb = 0.00062
    
    if velocity <= 0.:
        return 0., 0., 0., 0., 0.
    theta = np.arccos(direction[2]) #radians
    #sin_critical_angle = np.sqrt( 2*(energy-0.00062)/(4.002603254e9) )/(np.abs(velocity)/3e8)
    sin_critical_angle = np.sqrt(2.*mass*(energy - Eb))/momentum/1000. #Eq 2.19 from J. Adams thesis
    if sin_critical_angle > 1.:
        return 0., 0., 0., 0., 0.
    crit_angle = np.arcsin( sin_critical_angle ) #radians
    if theta > crit_angle:
        return 0., 0., 0., 0., 0.
    Velocity_He_atom=np.sqrt( 2.*(energy-Eb)/(mass) )*c
    
    #sin_theta_R = (velocity/c)*np.sin((theta))/np.sqrt( 2*(energy-Eb)/(mass) )
    sin_theta_R = (momentum/1000.)*np.sin(theta)/np.sqrt(2.*mass*(energy-Eb)) #Eq 2.18 from Adams thesis
    theta_R = np.arcsin(sin_theta_R)
    new_Vz=Velocity_He_atom*np.cos(theta_R)
    if direction[1] == 0.:
        tan = np.pi/2.
        new_Vy = 0.        
    else:
        tan = np.arctan(np.abs(direction[0]/direction[1]) )
        new_Vy= direction[1]/np.abs(direction[1])*Velocity_He_atom*np.sin(theta_R)*np.sin((tan))
    if direction[0] == 0.:
        new_Vx = 0.
    else:
        new_Vx= direction[0]/np.abs(direction[0])*Velocity_He_atom*np.sin(theta_R)*np.cos((tan))
        
    magnitude=np.sqrt(new_Vx*new_Vx+new_Vy*new_Vy+new_Vz*new_Vz)
    return 1., Velocity_He_atom, new_Vx/magnitude, new_Vy/magnitude, new_Vz/magnitude


def QP_propagation(start, conditions, reflection_prob, evap_eff=0.60):
    phi, arctheta = np.random.uniform(0., 2.*np.pi), np.random.uniform(-1., 1)
    theta = np.arccos(arctheta)
    dx = np.cos( phi ) * np.sin( theta )
    dy = np.sin( phi ) * np.sin( theta )
    dz = np.cos(theta)
    
    #dx, dy, dz = 0., 0., 1
    total_time = 0.
    n=0
    xs, ys, zs = [start[0]], [start[1]], [start[2]]
    momentum = Random_QPmomentum() #keV/c
    if momentum < 1.1:
        return 0, total_time, n, xs, ys, zs, momentum, None, -999
    velocity = QP_velocity(momentum) #m/s
    #print(velocity)
    energy = QP_dispersion(momentum) #eV
    evaporated = False
    while True:
        n+=1
        X1, Y1, Z1, surface_type = find_surface_intersection(start, [dx, dy, dz], conditions)
        if surface_type == None:
            return 0, total_time, n, xs, ys, zs, momentum, None, -999
        xs.append(X1)
        ys.append(Y1)
        zs.append(Z1)
        total_time += (np.sqrt(pow(X1-start[0],2.)+pow(Y1-start[1], 2.)+pow(Z1-start[2],2.)))/(velocity*1.0e-4) #us
        if surface_type == "Liquid":
            evap, velocity, dx, dy, dz = evaporation(momentum, energy, velocity, [dx, dy, dz])
            #evap = 1.
            #print("Liquid", velocity)
            #if evap < 0.5:
            if  (evap < 0.5) or (np.random.random() > evap_eff ):
                return 0, total_time, n, xs, ys, zs, momentum, surface_type, -999
            else:
                Z1 *= 1.005 #just above the liquid surface
                start = [X1, Y1, Z1]
                evaporated = True
                continue
            
        if "CPD" in surface_type:
            cpd_id = int(surface_type.split("CPD")[-1])
            if evaporated == True:
                return 1, total_time, n, xs, ys, zs, momentum, surface_type, cpd_id
            return 0, total_time, n, xs, ys, zs, momentum, surface_type, cpd_id
        r = np.random.random()
        if r > reflection_prob:
            return 0, total_time, n, xs, ys, zs, momentum, surface_type, -999
        if surface_type == "X":
            dx *= -1.
        if surface_type == "Y":
            dy *= -1.
        if surface_type == "Z":
            dz *= -1.
        if surface_type == "XY":
            normal_vector = np.array([X1, Y1, 0.])
            normal_vector = normal_vector / np.linalg.norm(normal_vector)
            incident_vector = np.array([dx, dy, dz])
            reflected_vector = incident_vector - 2. * np.dot(incident_vector, normal_vector) * normal_vector
            dx, dy, dz = reflected_vector / np.linalg.norm(reflected_vector)
        start = [X1, Y1, Z1]
        
        
def photon_propagation(start, conditions, reflection_prob):
    phi, arctheta = np.random.uniform(0., 2.*np.pi), np.random.uniform(-1., 1)
    theta = np.arccos(arctheta)
    dx = np.cos( phi ) * np.sin( theta )
    dy = np.sin( phi ) * np.sin( theta )
    dz = np.cos(theta)
    #dx, dy, dz = 0., 0., 1.
    total_time = np.random.exponential(0.004)
    n=0
    xs, ys, zs = [start[0]], [start[1]], [start[2]]
    velocity = 29979.2/1.03 #speed of light in He4 cm/us
    while True:
        n+=1
        X1, Y1, Z1, surface_type = find_surface_intersection(start, [dx, dy, dz], conditions)
        if surface_type == None:
            return 0,total_time, n, xs, ys, zs, -999
        xs.append(X1)
        ys.append(Y1)
        zs.append(Z1)
        total_time += np.sqrt(pow(X1-start[0],2.)+pow(Y1-start[1], 2.)+pow(Z1-start[2],2.))/velocity
        if "CPD" in surface_type:
            cpd_id = int(surface_type.split("CPD")[-1])
            return 1, total_time, n, xs, ys, zs, cpd_id
        r = np.random.random()
        if r > reflection_prob:
            return 0, total_time, n, xs, ys, zs, -999
        if surface_type == "X":
            dx *= -1.
        if surface_type == "Y":
            dy *= -1.
        if surface_type == "Z":
            dz *= -1.
        if surface_type == "XY":
            normal_vector = np.array([X1, Y1, 0.])
            normal_vector = normal_vector / np.linalg.norm(normal_vector)
            incident_vector = np.array([dx, dy, dz])
            reflected_vector = incident_vector - 2. * np.dot(incident_vector, normal_vector) * normal_vector
            dx, dy, dz = reflected_vector / np.linalg.norm(reflected_vector)
        start = [X1, Y1, Z1]

''' #########################################################################

    Define functions for getting the energy deposited in the CPDs
    
    ######################################################################### '''

def GetSingletSignal(detector, photons, X, Y, Z, useMap=True):
    '''
    Attempt to simulate the CPD response for singlet photons.
    
    If useMap == True, the detector's LCEmap is loaded and the number of photon hits are 
    calculated using that light collection efficiency given the interaction location.
    If False, loop through individual photons, calculating the number of hits 
    and the arrival times. This takes substantially more time: ~1.3s for 10k photons.
    Number of hits are converted to pulse areas using the Singlet Photon Energy
    '''
    #check to see if there's an LCEmap loaded/generated
    if type(detector.get_LCEmap()) == int:
        useMap = False
        
    nCPDs = detector.get_nCPDs()
    nHitsPerCPD = [0]*nCPDs
    arrivalTimes = []
    chAreas = [0.]*nCPDs
    #check to see if a map is loaded
    if useMap:
        hitProbabilities = detector.get_photon_hits( X, Y, Z)
        coincidence = 0
        for i in range(nCPDs):
            nHits = np.random.binomial(photons, hitProbabilities[i])
            coincidence += min([nHits, 1]) # 0 or 1
            chAreas[i] = nHits*Singlet_PhotonEnergy
            thisCPD = detector.get_CPD(i)
            noiseBaseline = thisCPD.get_baselineNoise()
            noise = np.random.normal(noiseBaseline[0], noiseBaseline[1], size=nHits)
            chAreas[i] += np.sum(noise)
            
        return CPD_Signal(sum(chAreas), chAreas, coincidence, arrivalTimes)
    
    #if the map hasn't been loaded or we want arrival times...
    conditions = detector.get_surface_conditions()
    for i in range(nCPDs):
        conditions.append( (detector.get_CPD(i)).get_surface_condition() )
    for n in range(photons):
        hit, arrival_time, n, xs, ys, zs, cpd_id = photon_propagation([X, Y, Z], conditions, detector.get_photon_reflection_prob())
        if hit > 0.5:
            nHitsPerCPD[cpd_id] += hit
            arrivalTimes.append( arrival_time )
    coincidence = 0
    #add in baseline noise for each detected photon
    for i in range(nCPDs):
        chAreas[i] = nHitsPerCPD[i]*Singlet_PhotonEnergy
        coincidence += min([nHitsPerCPD[i], 1]) # 0 or 1
        thisCPD = detector.get_CPD(i)
        noiseBaseline = thisCPD.get_baselineNoise()
        noise = np.random.normal(noiseBaseline[0], noiseBaseline[1], size=nHitsPerCPD[i])
        chAreas[i] += np.sum(noise)     
        
    return CPD_Signal(sum(chAreas), chAreas, coincidence, arrivalTimes)  



def GetEvaporationSignal(detector, QPs, X, Y, Z, useMap=True):
    '''
    Attempt to simulate the CPD response for quasiparticles.
    
    If useMap == True, the detector's QPEmap is loaded and the number of QP hits are 
    calculated using that evaporation efficiency given the interaction location.
    If False, loop through individual QPs, calculating the number of hits 
    and the arrival times. This takes substantially more time: ~1.5s for 10k particles.
    Number of hits are converted to pulse areas using the detector's adsorption gain. 
    
    '''
    #check to see if there's an LCEmap loaded/generated
    if type(detector.get_QPEmap()) == int:
        useMap = False
        
    nCPDs = detector.get_nCPDs()
    nHitsPerCPD = [0]*nCPDs
    arrivalTimes = []
    chAreas = [0.]*nCPDs
    #check to see if a map is loaded
    if useMap:
        hitProbabilities = detector.get_QP_hits( X, Y, Z)
        coincidence = 0    
        
        for i in range(nCPDs):
            nHits = np.random.binomial( QPs, hitProbabilities[i] )
            coincidence += min([nHits, 1]) # 0 or 1
            chAreas[i] = nHits * detector.get_adsorption_gain()
            thisCPD = detector.get_CPD(i)
            noiseBaseline = thisCPD.get_baselineNoise()
            noise = np.random.normal(noiseBaseline[0], noiseBaseline[1], size=nHits)
            chAreas[i] += np.sum(noise)
            
        return CPD_Signal(sum(chAreas), chAreas, coincidence, arrivalTimes)
    
    #if the map hasn't been loaded or we want arrival times...
    conditions = detector.get_surface_conditions()
    conditions.append( detector.liquid_surface )
    for i in range(nCPDs):
        conditions.append( (detector.get_CPD(i)).get_surface_condition() )
    for n in range(QPs):
        hit, arrival_time, n, xs, ys, zs, p, surf, cpd_id = QP_propagation([X, Y, Z], conditions, detector.get_QP_reflection_prob(), evap_eff=detector.get_evaporation_eff())
        if hit > 0.5:
            nHitsPerCPD[cpd_id] += hit
            arrivalTimes.append( arrival_time )
    
    coincidence = 0
    #add in baseline noise for each detected photon
    for i in range(nCPDs):
        chAreas[i] = nHitsPerCPD[i]*detector.get_adsorption_gain()
        coincidence += min([nHitsPerCPD[i], 1]) # 0 or 1
        thisCPD = detector.get_CPD(i)
        noiseBaseline = thisCPD.get_baselineNoise()
        noise = np.random.normal(noiseBaseline[0], noiseBaseline[1], size=nHitsPerCPD[i])
        chAreas[i] += np.sum(noise)     
        
    return CPD_Signal(sum(chAreas), chAreas, coincidence, arrivalTimes) 




import HeST.Detection as detection
import os
import numpy as np

#The detector geometry is defined from the point of view of particle paths.
# We essentially want to define various "surface conditions" where the particle paths are obstructed
# These functions also carry a "boundary_type", so that we can keep track if the particle is obstructed by
# a CPD, or a wall, and how it may reflect off of a given wall.

def sensor_conditions(x, y, z):
    boundary_type = "CPD0"
    radius = 3.8
    height = 3.3
    return (x*x + y*y < radius*radius) & (z < height) | (x*x + y*y >= radius*radius) , boundary_type

baseline_noise = [0., 0.]
phonon_conversion = 0.25
cpd = detection.VCPD(sensor_conditions, baseline_noise, phonon_conversion)


def wall_conditions(x, y, z):
    boundary_type = "XY"
    radius = 3. #cm
    height = 2.75 #cm
    return ((x*x + y*y < radius*radius) & (z < height) ) | (z > height), boundary_type

def bottom_conditions(x, y, z):
    boundary_type = "Z"
    bottom = 0. #cm
    return (z > bottom), boundary_type

def liquid_surface(x, y, z):
    boundary_type = "Liquid"
    height = 2.75 #cm
    return (z < height), boundary_type

def liquid_conditions(x, y, z):
    height = 2.75 #cm
    radius = 3. #cm
    bottom = 0. #cm
    return ((x*x + y*y < radius*radius) & (z < height) & (z > bottom))
    
DetectorExample_Amherst = detection.VDetector( [wall_conditions, bottom_conditions], liquid_surface=liquid_surface, liquid_conditions=liquid_conditions, adsorption_gain=6.0e-3, evaporation_eff=0.60, CPDs=[cpd] )
DetectorExample_Amherst.load_LCEmap(os.path.dirname(__file__)+'/LCE_map_amherstExample_41x41x30_noReflections.npy')
DetectorExample_Amherst.set_LCEmap_positions([np.linspace(-3., 3., 41), np.linspace(-3., 3., 41), np.linspace(0., 2.75, 30)])
DetectorExample_Amherst.load_QPEmap(os.path.dirname(__file__)+'/QP_map_amherstExample_41x41x30_noReflections.npy')
DetectorExample_Amherst.set_QPEmap_positions([np.linspace(-3., 3., 41), np.linspace(-3., 3., 41), np.linspace(0., 2.75, 30)])






import scipy.special as scispec
import numpy as np


#WIMP generator function taken from NEST, but pythonified


def WIMP_dRate(recoilEnergy_eV, mass_MeV):
    # We are going to hard code in the astrophysical halo for now.  This may be
    # something that we make an argument later, but this is good enough to start.
    # Some constants:

    ER = recoilEnergy_eV / 1000.
    mWimp = mass_MeV / 1000.
    V_SUN = 250.2     # +/- 1.4: arXiv:2105.00599, page 12 (V_EARTH 29.8km/s)
    V_WIMP = 238.     # +/- 1.5: page 12 and Table 1
    V_ESCAPE = 544.   # M.C. Smith et al. RAVE Survey
    RHO_NAUGHT = 0.3  # Local dm density in GeV/cc
    M_N = 0.9395654                  # Nucleon mass [GeV]
    N_A = 6.0221409e23              # Avogadro's number [atoms/mol]
    c = 2.99792458e10                # Speed of light [cm/s]
    GeVperAMU = 0.9315               # Conversion factor
    SecondsPerDay = 60. * 60. * 24.  # Conversion factor
    KiloGramsPerGram = 0.001         # Conversion factor
    keVperGeV = 1.0e6                 # Conversion factor
    cmPerkm = 1.0e5                   # Conversion factor
    SqrtPi = pow(np.pi, 0.5)
    root2 = np.sqrt(2.);
    #// Convert all velocities from km/s into cm/s
    v_0 = V_WIMP * cmPerkm;      # peak WIMP velocity
    v_esc = V_ESCAPE * cmPerkm;  # escape velocity
    v_e = (V_SUN + (0.49 * 29.8 * np.cos(-0.415 * 2. * np.pi))) * cmPerkm # the Earth's velocity

    #// Define the detector Z and A and the mass of the target nucleus
    Z = 2. #ATOM_NUM;
    A = 4. #Atom mass
    M_T = A * GeVperAMU;

    # Calculate the number of target nuclei per kg
    N_T = N_A / (A * KiloGramsPerGram);

    # Rescale the recoil energy and the inelastic scattering parameter into GeV
    ER /= keVperGeV;
    delta = 0. / keVperGeV #Setting this to a nonzero value will allow
    # for inelastic dark matter...
    # Set up your dummy WIMP model (this is just to make sure that the numbers
    # came out correctly for definite values of these parameters, the overall
    # normalization of this spectrum doesn't matter since we generate a definite
    # number of events from the macro).
    m_d = mWimp       # [GeV]
    sigma_n = 1.0e-36  #[cm^2] 1 pb reference

    #// Calculate the other factors in this expression
    mu_ND = mWimp * M_N / (mWimp + M_N) #WIMP-nucleON reduced mass
    mu_TD = mWimp * M_T / (mWimp + M_T) #WIMP-nucleUS reduced mass
    fp = 1.  #// Neutron and proton coupling constants for WIMP interactions.
    fn = 1.

    # Calculate the minimum velocity required to give a WIMP with energy ER
    v_min = 0.
    if (ER != 0.):
        v_min = c * (((M_T * ER) / mu_TD) + delta) / (root2 * np.sqrt(M_T * ER))

    bet = 1.

    # Start calculating the differential rate for this energy bin, starting
    # with the velocity integral:
    x_min = v_min / v_0 # Use v_0 to rescale the other velocities
    x_e = v_e / v_0
    x_esc = v_esc / v_0

    # Calculate overall normalization to the velocity integral
    N = SqrtPi * SqrtPi * SqrtPi * v_0 * v_0 * v_0
    N *= (scispec.erf(x_esc) - (4. / SqrtPi) * np.exp(-x_esc * x_esc) * (x_esc / 2. + bet * x_esc * x_esc * x_esc / 3.))
    # Calculate the part of the velocity integral that isn't a constant
    zeta = 0.
    thisCase = -1
    if ((x_e + x_min) < x_esc):
        zeta = ((SqrtPi * SqrtPi * SqrtPi * v_0 * v_0) / (2. * N * x_e))
        zeta *= (scispec.erf(x_min + x_e) - scispec.erf(x_min - x_e) - ((4. * x_e) / SqrtPi) * np.exp(-x_esc * x_esc) * (1 + bet * (x_esc * x_esc - x_e * x_e / 3. - x_min * x_min)))

    if ((x_min > abs(x_esc - x_e)) and ((x_e + x_esc) > x_min)):
        zeta = ((SqrtPi * SqrtPi * SqrtPi * v_0 * v_0) / (2. * N * x_e))
        zeta *= (scispec.erf(x_esc) + scispec.erf(x_e - x_min) - (2. / SqrtPi) * np.exp(-x_esc * x_esc) * (x_esc + x_e - x_min - (bet / 3.) * (x_e - 2. * x_esc - x_min) * (x_esc + x_e - x_min) * (x_esc + x_e - x_min)))

    if (x_e > (x_min + x_esc)):
        zeta = 1. / (x_e * v_0)

    if ((x_e + x_esc) < x_min):
        zeta = 0.


    a = 0.52                           # in fm
    C = 1.23 * pow(A, 1. / 3.) - 0.60  # fm
    s = 0.9  # skin depth of nucleus in fm. Originally used by Karen Gibson; XENON100 1fm; 2.30 acc. to Lewin and Smith maybe?
    rn = np.sqrt(C * C + (7. / 3.) * np.pi * np.pi * a * a - 5. * s * s);
    ''' alternatives: 1.14*A^1/3 given in L&S, or
        rv=1.2*A^1/3 then rn =
        sqrt(pow(rv,2.)-5.*pow(s,2.)); used by
        XENON100 (fm)
    '''
    q = 6.92 * np.sqrt(A * ER) # in units of 1 over distance or length
    if (q * rn > 0.):
        FormFactor = 3. * np.exp(-0.5 * q * q * s * s) * (np.sin(q * rn) - q * rn * np.cos(q * rn))
        FormFactor /= (q * rn * q * rn * q * rn) # qr and qs unitless inside Bessel function, which is dimensionless too
    else:
        FormFactor = 1.;

    # Now, the differential spectrum for this bin!
    dSpec = 0.5 * (c * c) * N_T * (RHO_NAUGHT / m_d) * (M_T * sigma_n / (mu_ND * mu_ND));
    # zeta=1.069-1.4198*ER+.81058*pow(ER,2.)-.2521*pow(ER,3.)+.044466*pow(ER,4.)-0.0041148*pow(ER,5.)+0.00013957*pow(ER,6.)+2.103e-6*pow(ER,7.);
    # if ( ER > 4.36 ) squiggle = 0.; //parameterization for 7 GeV WIMP using
    # microMegas
    dSpec *= (((Z * fp) + ((A - Z) * fn)) / fn) * (((Z * fp) + ((A - Z) * fn)) / fn) * zeta * FormFactor * FormFactor * SecondsPerDay / keVperGeV;

    return dSpec;


def WIMP_spectrum_prep(mass_MeV, minE = 21):
    test_e = np.geomspace(minE, 10000, 100)
    f = []
    for i in test_e:
        f.append( WIMP_dRate(i, mass_MeV) )
    f = np.array(f)

    cut_off = (f > f[0]/1e7)
    maxE = test_e[cut_off][-1]
    return test_e[0], maxE, f[0]

def WIMP_spectrum(mass_MeV, minE_eV, maxE_eV, maxY):

    maxE = np.log10(maxE_eV * 1.2)
    minE = np.log10(minE_eV)
    #print( minE, maxE)
    while True:
        yTry = np.random.random()*maxY
        xTry = pow(10., np.random.uniform(minE, maxE) )
        if (yTry < WIMP_dRate(xTry, mass_MeV)):
            return xTry



import HeST.Detection as detection
import os
import numpy as np

#The detector geometry is defined from the point of view of particle paths.
# We essentially want to define various "surface conditions" where the particle paths are obstructed
# These functions also carry a "boundary_type", so that we can keep track if the particle is obstructed by
# a CPD, or a wall, and how it may reflect off of a given wall.
def sensor1_conditions(x, y, z):
    boundary_type = "CPD0"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD1 = ( (x < CPD_length/2.0 + CPD_XYpitch/2.0) & (x > -CPD_length/2.0 + CPD_XYpitch/2.0) & (y < CPD_length/2.0 + CPD_XYpitch/2.0) & (y > -CPD_length/2.0 + CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset + CPDthickness/2.0) )
    return ~CPD1, boundary_type
    
def sensor2_conditions(x, y, z):
    boundary_type = "CPD1"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD2 = ( (x < CPD_length/2.0 - CPD_XYpitch/2.0) & (x > -CPD_length/2.0 - CPD_XYpitch/2.0) & (y < CPD_length/2.0 + CPD_XYpitch/2.0) & (y > -CPD_length/2.0 + CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset + CPDthickness/2.0) )
    return ~CPD2, boundary_type
    
def sensor3_conditions(x, y, z):
    boundary_type = "CPD2"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD3 = ( (x < CPD_length/2.0 + CPD_XYpitch/2.0) & (x > -CPD_length/2.0 + CPD_XYpitch/2.0) & (y < CPD_length/2.0 - CPD_XYpitch/2.0) & (y > -CPD_length/2.0 - CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset + CPDthickness/2.0) )
    return ~CPD3, boundary_type
    
def sensor4_conditions(x, y, z):
    boundary_type = "CPD3"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD4 = ( (x < CPD_length/2.0 - CPD_XYpitch/2.0) & (x > -CPD_length/2.0 - CPD_XYpitch/2.0) & (y < CPD_length/2.0 - CPD_XYpitch/2.0) & (y > -CPD_length/2.0 - CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset + CPDthickness/2.0) )
    return ~CPD4, boundary_type
    
def sensor5_conditions(x, y, z):
    boundary_type = "CPD4"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD5 = ( (x < CPD_length/2.0 + CPD_XYpitch/2.0) & (x > -CPD_length/2.0 + CPD_XYpitch/2.0) & (y < CPD_length/2.0 + CPD_XYpitch/2.0) & (y > -CPD_length/2.0 + CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPD_Zpitch - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset - CPD_Zpitch + CPDthickness/2.0) )
    return ~CPD5, boundary_type
    
def sensor6_conditions(x, y, z):
    boundary_type = "CPD5"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD6 = ( (x < CPD_length/2.0 - CPD_XYpitch/2.0) & (x > -CPD_length/2.0 - CPD_XYpitch/2.0) & (y < CPD_length/2.0 + CPD_XYpitch/2.0) & (y > -CPD_length/2.0 + CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPD_Zpitch - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset - CPD_Zpitch + CPDthickness/2.0) )
    return ~CPD6, boundary_type
    
def sensor7_conditions(x, y, z):
    boundary_type = "CPD6"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD7 = ( (x < CPD_length/2.0 + CPD_XYpitch/2.0) & (x > -CPD_length/2.0 + CPD_XYpitch/2.0) & (y < CPD_length/2.0 - CPD_XYpitch/2.0) & (y > -CPD_length/2.0 - CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPD_Zpitch - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset - CPD_Zpitch + CPDthickness/2.0) )
    return ~CPD7, boundary_type
    
def sensor8_conditions(x, y, z):
    boundary_type = "CPD7"
    CPDthickness = .1 #cm
    CPD_length = 1 # cm
    det_Zoffset = 3.0 #cm
    cal_Zoffset = -3.8743 #cm
    CPD_XYpitch = 1.373 #cm
    CPD_Zpitch = 3.878 #cm
    CPD8 = ( (x < CPD_length/2.0 - CPD_XYpitch/2.0) & (x > -CPD_length/2.0 - CPD_XYpitch/2.0) & (y < CPD_length/2.0 - CPD_XYpitch/2.0) & (y > -CPD_length/2.0 - CPD_XYpitch/2.0) & (z > det_Zoffset + cal_Zoffset - CPD_Zpitch - CPDthickness/2.0) & (z < det_Zoffset + cal_Zoffset - CPD_Zpitch + CPDthickness/2.0) )
    return ~CPD8, boundary_type
    
    

baseline_noise = [0., 0.]
phonon_conversion = 0.25
cpd1 = detection.VCPD(sensor1_conditions, baseline_noise, phonon_conversion)
cpd2 = detection.VCPD(sensor2_conditions, baseline_noise, phonon_conversion)
cpd3 = detection.VCPD(sensor3_conditions, baseline_noise, phonon_conversion)
cpd4 = detection.VCPD(sensor4_conditions, baseline_noise, phonon_conversion)
cpd5 = detection.VCPD(sensor5_conditions, baseline_noise, phonon_conversion)
cpd6 = detection.VCPD(sensor6_conditions, baseline_noise, phonon_conversion)
cpd7 = detection.VCPD(sensor7_conditions, baseline_noise, phonon_conversion)
cpd8 = detection.VCPD(sensor8_conditions, baseline_noise, phonon_conversion)

def wall_conditions(x, y, z):
    boundary_type = "XY"
    radius = 2.381 #cm
    bottomPos = -8.407 #cm
    topPos = -2.791 #cm
    det_Zoffset = 3.0 #cm
    Z_correction = 0 #cm
    return ~((x*x + y*y > radius*radius)) , boundary_type

def top_conditions(x,y,z):
    boundary_type = "Z"
    radius = 2.381 #cm
    bottomPos = -8.407 #cm
    topPos = -2.791 #cm
    det_Zoffset = 3.0 #cm
    Z_correction = 0 #cm
    return z < topPos + det_Zoffset + Z_correction, boundary_type

def bottom_conditions(x, y, z):
    boundary_type = "Z"
    radius = 2.381 #cm
    bottomPos = -8.407 #cm
    topPos = -2.791 #cm
    det_Zoffset = 3.0 #cm
    Z_correction = 0 #cm
    return z > bottomPos + det_Zoffset + Z_correction, boundary_type


def liquid_surface(x, y, z):
    boundary_type = "Liquid"
    height = 2.75 #cm
    bottomPos = -8.407 #cm
    topPos = -2.791 #cm
    det_Zoffset = 3.0 #cm
    Z_correction = 0 #cm
    return (z < bottomPos + det_Zoffset + Z_correction + height), boundary_type

def liquid_conditions(x, y, z):
    height = 2.75 #cm
    radius = 2.381 #cm
    bottomPos = -8.407 #cm
    topPos = -2.791 #cm
    det_Zoffset = 3.0 #cm
    Z_correction = 0 #cm
    return ((x*x + y*y < radius*radius) & (z < bottomPos + det_Zoffset + Z_correction + height) & (z > bottomPos + det_Zoffset + Z_correction))

#create the detector using the conditions above
DetectorExample_LBNL = detection.VDetector( [wall_conditions, top_conditions, bottom_conditions], liquid_surface=liquid_surface, liquid_conditions=liquid_conditions,
                                adsorption_gain=6.0, evaporation_eff=0.60, CPDs=[cpd1, cpd2, cpd3, cpd4, cpd5, cpd6, cpd7, cpd8],
                                photon_reflection_prob=0., QP_reflection_prob=0.)





import scipy.special as scispec
import numpy as np

#Global constants
Singlet_PhotonEnergy = 15.5 #eV
Triplet_PhotonEnergy = 18.0 #eV

#create some other classes to store results
#unlike NEST, we won't have a mean yields object; just the quanta object with fluctuations applied
class QuantaResult:
    def __init__(self, SingletPhotons, TripletPhotons, Quasiparticles):
        self.SingletPhotons = SingletPhotons
        self.TripletPhotons = TripletPhotons
        self.Quasiparticles = Quasiparticles

class CPD_Signal:
    def __init__(self, area_eV, chArea_eV, coincidence, arrivalTimes_us=[]):
        self.area_eV = area_eV #total pulse area
        self.chArea_eV = chArea_eV # individual CPD pulse areas
        self.coincidence = coincidence #number of CPDs hit
        self.arrivalTimes_us = arrivalTimes_us #time at which the particle hits the CPD

# Polynomial functions to get the energy channel partitioning for ERs and NRs

def ER_QP_eFraction(energy):
    a, b, c, d = 100., 2.98839372, 13.89437102, 0.33504361
    frac =  a/((energy-c)**b) + d
    if frac > 1.0 or energy < 19.77:
        return 1.0
    return frac

def ER_triplet_eFraction(energy):
    a, b, c, d, e, f, g = 14.69346932, 1.17858385, 17.60481951, 0.95286617, 0.23929604, 13.81618099, 4.92551215
    frac = f/(energy-a)**b - g/(energy-c)**d + e
    if frac < 0. or energy < 19.77:
        return 0.
    return frac

def ER_singlet_eFraction(energy):
    a, b, c, d = 2.05492076e+03, 5.93097885, 3.3091563, 0.31773768
    frac = -a/(energy-b)**c + d
    if frac < 0. or energy < 19.77:
        return 0.
    return frac

def NR_QP_eFraction(energy):
    logE = np.log10(energy)
    a, b, c, d, e, f, g =  1.37928294, 1.62782641, -3.64361207, 2.25802903, 0.07210282, -0.31126613, -0.00495869
    frac = a/logE/logE + b/logE + c + d*logE + e*logE*logE + f*logE*logE*logE + \
           e*logE*logE*logE*logE + g*logE*logE*logE*logE*logE
    if frac > 1.0 or energy < 19.77:
        return 1.0
    return frac

def NR_triplet_eFraction(energy):
    x = np.log10(energy)
    a, b, c, d, e, f, g, h =  -7.73003451, 17.1117164, -12.0306409, 4.388470, -2.5361635, 0.661017, -8.41417e-2,  4.2301e-03
    frac =  a/x + b + c*x + d*x*x*x + e*x*x*x*x + f*x*x*x*x*x + g*x*x*x*x*x*x + h*x*x*x*x*x*x*x
    if frac < 0 or energy < 19.77:
        return 0.
    return frac

def NR_singlet_eFraction(energy):
    x = np.log10(energy)
    a, b, c, d = -6.74959165e+01,  1.70997665e+02, -1.42571001e+02,  8.45017681e+01
    e, f, g, h = -6.92781245e+01,  2.87318953e+01, -7.05306162e+00,  1.03483204e+00
    i, j  = -8.39958299e-02,  2.90462675e-03
    frac = a/x + b + c*x + d*x*x*x + e*x*x*x*x + f*x*x*x*x*x + g*x*x*x*x*x*x + h*x*x*x*x*x*x*x + i*x*x*x*x*x*x*x*x + j*x*x*x*x*x*x*x*x*x
    if frac < 0. or energy < 19.77:
        return 0.
    return frac


#combine the above functions into a single function
def GetEnergyChannelFractions(energy, interaction):
    #returns the mean fraction of energy in
    # the singlet, triplet, QP, and IR channels
    energy = min([energy, 1.0e5])
    # energy -- recoil energy in eV
    if interaction == "ER":
        singlet = ER_singlet_eFraction(energy)
        triplet = ER_triplet_eFraction(energy)
        QP      = ER_QP_eFraction(energy)
        IR      = max([0., energy-singlet-triplet-QP])
    else:
        if interaction == "NR":
            singlet = NR_singlet_eFraction(energy)
            triplet = NR_triplet_eFraction(energy)
            QP      = NR_QP_eFraction(energy)
            IR      = max([0., energy-singlet-triplet-QP])
        else:
            print("Please specify ER or NR for interaction type!")
            return 1

    return singlet, triplet, QP, IR

#HeST main yields function
def GetSingletYields(energy, interaction):
    # returns the mean prompt scintillation yield for
    # the prompt + exponential + 1/t components

    # energy -- recoil energy in eV
    # interaction -- "ER" or "NR"

    detection_gain = 0.052 #to get the LBL yields to match with the ~35% energy fraction in prompt yields

    meanERyield = 1.25049e-3# +/- 0.03153 phe/eV
    meanNRyield = 0.477545e-3 # +/- 0.006395 phe/eV

    ERsingletFraction = 0.85824614
    NRsingletFraction = 0.72643866
    if energy < 19.77:
        return 0.
    if interaction == "ER":
        return meanERyield*energy/detection_gain * ERsingletFraction #convert from phe/keV to n_photons
    else:
        if interaction == "NR":
            return meanNRyield*energy/detection_gain *NRsingletFraction #convert from phe/keV to n_photons
        else:
            print("Please specify ER or NR for interaction type!")

def Get_Quasiparticles(qp_energy): 
    #returns the mean number of quasiparticles given 
    # the energy in eV in the QP channel
    
    #based on a linear fit after calculating the nQPs
    # using random sampling of the dispersion relation
    # given a Bose-Einstein distribution for nD/dp
    
    slope_t2, b_t2 = 1244.04521473, 29.10987933 #linear fit params
    return slope_t2*qp_energy + b_t2

def GetQuanta(energy, interaction):
    # energy -- recoil energy in eV
    # interaction -- "ER" or "NR"
 
    singlet_fraction, triplet_fraction, QP_fraction, IR_fraction = GetEnergyChannelFractions(energy, interaction)
    nSingletPhotons = GetSingletYields(energy, interaction)
    singlet_energy = nSingletPhotons * Singlet_PhotonEnergy
    Fano = 1.0
    if singlet_fraction > 0.:
        triplet_energy = (triplet_fraction/singlet_fraction) * singlet_energy
        QP_energy = (QP_fraction/singlet_fraction) * singlet_energy
    
        nTripletPhotons = triplet_energy / Triplet_PhotonEnergy #(eV / (eV/ph))
    
        scint_energy = singlet_energy + triplet_energy
    
        nPhotons = nSingletPhotons + nTripletPhotons
    
        nSing_actual = max([0, int( np.random.normal(nSingletPhotons, np.sqrt(Fano*nSingletPhotons)) )])
        nTrip_actual = max([0, int( np.random.normal(nTripletPhotons, np.sqrt(Fano*nTripletPhotons)) )])

        scint_energy_actual = nSing_actual * Singlet_PhotonEnergy + nTrip_actual * Triplet_PhotonEnergy
    
        #assume the fano fluctuations are anti-correlated with the QP energy
        QP_energy += scint_energy - scint_energy_actual 
        nQP_actual = max([0, int(Get_Quasiparticles(QP_energy))])
    else:
        nSing_actual = 0
        nTrip_actual = 0
        QP_energy = energy
        nQP = int(Get_Quasiparticles(QP_energy))
        nQP_actual = max([0, int( np.random.normal(nQP, np.sqrt(Fano*nQP)) )])
    
    return QuantaResult( nSing_actual, nTrip_actual, nQP_actual )



# Quasiparticle functions

def QP_dispersion(p):
    params = [-2.26246976e-03, 9.13297690e-01, -9.43774817e-01, 2.81293178, -3.64233027, \
              2.38490219, -8.76986870e-01, 1.82682488e-01, -2.00519847e-02, 8.99140180e-04]
    e = 0.
    for i in range(len(params)):
        e += params[i]*pow(p, i)
    return e/1000. #returns energy in eV

def QP_velocity(p):
    params = [236.47194598487116, 142.7789374022186, -760.901788379717, 2197.6009596586828, \
              -3370.919165124788, 2734.173890494752, -1262.64114986808, 341.97960568115644,\
              -53.286858540286154, 4.369764313313597, -0.14355574751607653]

    v = 0.
    for i in range(len(params)):
        v += params[i]*pow(p, i)
    return v


def getMaxY_Temp(T):
    p = np.linspace(0, 5, 50)
    f = p*p/(np.exp(QP_dispersion(p)/(T/11606.)) -1.)
    return  max(f)*1.05


def Random_QPmomentum(T=2., maxY=0.2):
    while True:
        pTry = np.random.uniform(0., 4.7)
        yTry = np.random.uniform(0., maxY)
        if yTry <  pTry*pTry/(np.exp(QP_dispersion(pTry)/(T/11606.)) -1.):
            return pTry

def generate_quasiparticles(energy, T=2.):
    #make test draws for momentum and probability
    p = np.linspace(0, 4.7, 5000) #momentum range to consider

    f = p*p/(np.exp(QP_dispersion(p)/(T/11606.)) -1.)

    c = (f > 0.) #make sure there are no zero or negative values
    p, f = p[c], f[c]
    sumf = sum(f)
    f = f/sumf #normalize probabilities to unity

    #get the number of expected draws based on total energy
    expectedDraws = int(energy/0.0005 )
    initDraws = np.random.choice(p, p=f, size=expectedDraws)
    initEnergies = QP_dispersion(initDraws)
    QP_energies = initEnergies
    eSum = sum(initEnergies)
    nQP = expectedDraws

    if eSum > energy:
        cumEnergies = np.cumsum(QP_energies)
        truncate = ( cumEnergies < energy )
        truncEnergies = QP_energies[truncate]
        return len(truncEnergies)#, truncEnergies

    #QP_energies = list(QP_energies)
    #fill in the rest of the energies
    while eSum < energy:
        q = QP_dispersion(np.random.choice(p, p=f, size=1)[0])
        eSum += q
        nQP += 1
        #QP_energies.append( q )
    return nQP#, np.array(QP_energies)






from setuptools import setup, find_packages

setup(
    name='HeST',
    version='0.1.2',
    packages=find_packages(),
    install_requires=[ 'numpy', 'scipy'],
    package_data={
        'HeST': ['*.npy'],
    }
)
